import 'dart:async';
import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'variables.dart';
import 'api.dart';
import 'dart:developer';
import 'function.dart';


class ChangeEvent {
  final EventType type;
  final Node? oldNode;
  final Node? newNode;

  ChangeEvent(this.type, this.oldNode, this.newNode);

  @override
  String toString() {
    return 'ChangeEvent{type: $type, oldNode: $oldNode, newNode: $newNode}';
  }
}


class _ConditionalCompleter {
  final bool Function(List<ChangeEvent>) canComplete;
  final List<ChangeEvent> Function(List<ChangeEvent>) createResult;

  final _completer = Completer<List<ChangeEvent>>();

  Future<List<ChangeEvent>> get future => _completer.future;

  _ConditionalCompleter(this.canComplete, this.createResult);

  void complete(List<ChangeEvent> events) {
    if (!canComplete(events)) return;
    _completer.complete(createResult(events));
  }
}


class SensorValueListener with StorageListener {
  List<ChangeEvent> events = <ChangeEvent>[];
  final Set<_ConditionalCompleter> completers = {};
  var task = [];
  var mEvents = [];
  var uEvents = [];
  var dEvents = [];

  @override
  void gotStorageChange(EventType event, Node? oldNode, Node? newNode) async {
    //Event e = Event(event, oldNode, newNode);
 //   ChangeEvent e = ChangeEvent(event, oldNode, newNode);
//    events.add(ChangeEvent(event, oldNode, newNode));
//    for (final completer in completers) {
//      completer.complete(events);
//    }
      log('indiC new event ${newNode!.owner}');

  if(newNode.path.contains(':data:metrics') && newNode.name != 'metrics' ){

        log('============================got event metrics newNode ${newNode.path}');
        //await Future.delayed(const Duration(milliseconds: 1000));
        Storage.getValuePath.add(newNode.parentPath);
        await updateData(newNode);

  }else if(newNode.path.contains(':data:GeigerScoreUser') ){


          var str = await newNode.getValue('implementedRecommendations');

          if(str!.getValue('en') != CoreValues.uuidUserImplementedRecommendations.toString().replaceAll('[','').replaceAll(']','').trim()) {
            log('============================got event GeigerScoreUser newNode ${newNode.path}');
            //log('geiger-indicator - new IR: '+ str.getValue('en')!);
            //log('geiger-indicator - old IR: '+ CoreValues.uuidUserImplementedRecommendations.toString().replaceAll('[','').replaceAll(']','').trim());
            await generateImplementedReco(str.getValue('en'),Types.users);
            log('indiA User reco ${CoreValues.uuidUserImplementedRecommendations}');
            await updateScoreNode(Types.users, Types.geigerScoreUser);
            await aggregateScore();
          }


  }else if(newNode.path.contains(':data:GeigerScoreDevice') ){



  log('============================got event GeigerScoreDevice newNode ${newNode.path}');

  //await Future.delayed(const Duration(milliseconds: 2000));
      var str = await newNode.getValue('implementedRecommendations');

    if(str!.getValue('en') != CoreValues.uuidDeviceImplementedRecommendations.toString().replaceAll('[','').replaceAll(']','').trim()){
     // log('geiger-indicator - new IR: '+ str.getValue('en')!);
     // log('geiger-indicator - old IR: '+ CoreValues.uuidDeviceImplementedRecommendations.toString().replaceAll('[','').replaceAll(']','').trim());
      await generateImplementedReco(str.getValue('en'),Types.devices);
      log('indiA Device reco ${CoreValues.uuidDeviceImplementedRecommendations}');
      await updateScoreNode(Types.devices, Types.geigerScoreDevice);
      await aggregateScore();
    }

  }

/*
    if(newNode!.path.contains(':data:metrics') && newNode.name != 'metrics'){
      log('indiA got Sensor data from Plugin ${newNode}');


      if(mEvents.isEmpty){
          mEvents.add(newNode);
      }else if(!mEvents.contains(newNode)){
        mEvents.add(newNode);
      }


      if(mEvents.isNotEmpty){
        log('============================got event newNode ${newNode.path}');
        log('indiA merticsList $mEvents');
        //await Future.delayed(const Duration(milliseconds: 1000));
        Storage.getValuePath.add(newNode.parentPath);
        await updateData(events);
        log('indiA after merticsList $mEvents');
      }

    }else if(newNode.path.contains(':data:GeigerScoreUser')){
      log('indiC in GeigerScoreUser' );

      if(uEvents.isEmpty){
          uEvents.add(newNode);
          //log('indiC add $uEvents' );
      }else if(!uEvents.contains(newNode)){
          //log('indiC ureco List $uEvents' );
          //log('indiC add $newNode' );
          uEvents.add(newNode);
      }


      if(uEvents.isNotEmpty){

          //log('============================got event GeigerScoreUser newNode ${newNode.path}');
          //log('indiA ureco List $uEvents' );
          //await Future.delayed(const Duration(milliseconds: 2000));
          var str = await newNode.getValue('implementedRecommendations');

          if(str!.getValue('en') != CoreValues.uuidUserImplementedRecommendations.toString().replaceAll('[','').replaceAll(']','').trim()) {
            log('============================got event GeigerScoreUser newNode ${newNode.path}');
            log('geiger-indicator - new IR: '+ str.getValue('en')!);
            log('geiger-indicator - old IR: '+ CoreValues.uuidUserImplementedRecommendations.toString().replaceAll('[','').replaceAll(']','').trim());
            //log('indiC change');
            await generateImplementedReco(str.getValue('en'),Types.users);
            log('indiA User reco ${CoreValues.uuidUserImplementedRecommendations}');
            await updateScoreNode(Types.users, Types.geigerScoreUser);
          }
          uEvents.remove(newNode);
          log('indiA after ureco List $uEvents' );
      }
      log('indiC out GeigerScoreUser' );

    }else if(newNode.path.contains(':data:GeigerScoreDevice') ){
      log('indiC in GeigerScoreDevice $dEvents' );

      if(dEvents.isEmpty){
        log('indiC add $dEvents' );
        dEvents.add(newNode);
      }else{
          if(!dEvents.contains(newNode)){
            log('indiC dreco List $dEvents' );
            log('indiC add $newNode' );
            dEvents.add(newNode);
          }
      }


      if(dEvents.isNotEmpty){
          log('============================got event newNode ${newNode.path}');
          log('indiA dreco List $dEvents' );
          //await Future.delayed(const Duration(milliseconds: 2000));
          var str = await newNode.getValue('implementedRecommendations');
          log('geiger-indicator - new IR: '+ str!.getValue('en')!);
          log('geiger-indicator - old IR: '+ CoreValues.uuidDeviceImplementedRecommendations.toString().replaceAll('[','').replaceAll(']','').trim());

          if(str.getValue('en') != CoreValues.uuidDeviceImplementedRecommendations.toString().replaceAll('[','').replaceAll(']','').trim()){
          generateImplementedReco(str.getValue('en'),Types.devices);
          await updateScoreNode(Types.devices, Types.geigerScoreDevice);

          }
          dEvents.remove(newNode);
          log('indiA after dreco List $dEvents' );
      }
      log('indiC out GeigerScoreDevice' );
    }

    //if(newNode?.name != 'metrics') await updateData(newNode!);

*/

  }

//  Future<List<ChangeEvent>> getNumEvents(int num, [int timeout = 2000]) async {
//    int start = DateTime.now().millisecondsSinceEpoch;
//    List<ChangeEvent> ret = await Future.doWhile(() =>
//    events.length < num &&
//        start + 1000 * timeout > DateTime.now().millisecondsSinceEpoch)
//        .then((value) => events);
//
//    if (events.length < num) {
//      throw TimeoutException('Timeout reached while waiting for $num events');
//    }
//    return ret;
//  }

  Future<List<ChangeEvent>> awaitCount(int count,
      [Duration timeLimit = const Duration(seconds: 1)]) {
    canComplete(List events) => events.length >= count;
    createResult(List<ChangeEvent> events) =>
        List<ChangeEvent>.from(events.getRange(0, count));
    if (canComplete(events)) {
      return Future.value(createResult(events));
    }

    final completer = _ConditionalCompleter(canComplete, createResult);
    completers.add(completer);
    return completer.future.timeout(timeLimit, onTimeout: () {
      completers.remove(completer);
      throw TimeoutException(
          'Did not receive enough change events before timeout.', timeLimit);
    });
  }
}


Future aggregateScore()async{
  //await getDevices();
  await getScores();
  await getEmployees();
  await crtAggregateScore();
  await updateScoreNode(Types.users, Types.geigerScoreAggregate);
}


Future updateData(Node? n) async{

  try {
    //while (iterator.moveNext()) {
      //var node = iterator.current.newNode;
    //for(Node? n in e) {


      var temp = n!.path.toString().split(':');

      var type = '';
      temp[1] == Types.users ? type = Types.users : type = Types.devices;

      var type2 = '';
      temp[1] == Types.users ? type2 = Types.geigerScoreUser : type2 =
      Types.geigerScoreDevice;

//      log((await n.getValue('GEIGERvalue')).toString());

      if((await n.getValue('GEIGERvalue')) != null) {

        //await clearPN();
        await updateMetrics(type, n);
        await updateScoreNode(type, type2);
        await updateRecommendationNode(type);

       await aggregateScore();

      }else{
        log('indiA updateMetrics $n GEIGERvalue is not existed');
        return;
      }

      //e.remove(n);

  //}
//}
}catch(e){
    log('indiA updateData ${e.toString()}');
    //print(e);
  }

}


